/* jpgcom.c:  print any comments present in the specified JPEG image files

   This program is a heavily-stripped version of the "rdjpgcom" program
   distributed with version 6b of the Indendent JPEG Group's software.  The
   original version (along with a copy of the Indendent JPEG Group's README
   file) can be found in the directory orig/.

   The authors make NO WARRANTY or representation, either express or implied,
   with respect to this software, its quality, accuracy, merchantability, or
   fitness for a particular purpose.  This software is provided "AS IS", and
   you, its user, assume the entire risk as to its quality and accuracy.

   This software is copyright (C) 1991-1998, Thomas G. Lane.
   All Rights Reserved except as specified below.

   Permission is hereby granted to use, copy, modify, and distribute this
   software (or portions thereof) for any purpose, without fee, subject to
   these conditions:
   (1) If any part of the source code for this software is distributed, then
   this README file must be included, with this copyright and no-warranty
   notice unaltered; and any additions, deletions, or changes to the original
   files must be clearly indicated in accompanying documentation.
   (2) If only executable code is distributed, then the accompanying
   documentation must state that "this software is based in part on the work of
   the Independent JPEG Group".
   (3) Permission for use of this software is granted only if the user accepts
   full responsibility for any undesirable consequences; the authors accept NO
   LIABILITY for damages of any kind.

   Modifications Copyright (C)2004-2005 by Brian Lindholm.  All rights reserved
   except as specified above.  */


#include "../config.h"
#include <ctype.h>              /* to declare isupper(), tolower() */
#include <limits.h>             /* to determine PATH_MAX */
#include <stdio.h>              /* other I/O functions */
#include <stdlib.h>             /* other I/O functions */
#include <string.h>             /* other I/O functions */

#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#define OPTEND -1
#else
#define OPTEND EOF
#endif

#ifdef __MINGW32__
extern int getopt (int argc, char * const *argv, const char *optstring);
extern char *optarg;
extern int optind;
#endif

#ifndef PATH_MAX
#define PATH_MAX 256
#endif

/*
 * These macros are used to read the input file.  To reuse this code in another
 * application, you might need to change these.
 */

static FILE *jpgfile;            /* input JPEG file */

/* Return next input byte, or EOF if no more */
#define NEXTBYTE() getc(jpgfile)

/* Error handler */
#define ERR_RET(msg) { fprintf(stderr, " %s", msg); return(EOF); }


/* Read one byte, testing for EOF */
static int
read_1_byte (void)
{
  int c;

  c = NEXTBYTE ();
  if (c == EOF)
    ERR_RET ("Premature-EOF");
  return (c);
}


/* Read 2 bytes, convert to unsigned int */
/* All 2-byte quantities in JPEG markers are MSB first */
static int
read_2_bytes (void)
{
  int c1, c2;

  c1 = NEXTBYTE ();
  if (c1 == EOF)
    ERR_RET ("Premature-EOF");
  c2 = NEXTBYTE ();
  if (c2 == EOF)
    ERR_RET ("Premature-EOF");
  return ((c1 << 8) + c2);
}


/*
 * JPEG markers consist of one or more 0xFF bytes, followed by a marker code
 * byte (which is not an FF).  Here are the marker codes of interest in this
 * program.
 */

#define M_SOF0  0xC0            /* Start Of Frame N */
#define M_SOF1  0xC1            /* N indicates which compression process */
#define M_SOF2  0xC2            /* Only SOF0-SOF2 are now in common use */
#define M_SOF3  0xC3
#define M_SOF5  0xC5            /* NB: codes C4 and CC are NOT SOF markers */
#define M_SOF6  0xC6
#define M_SOF7  0xC7
#define M_SOF9  0xC9
#define M_SOF10 0xCA
#define M_SOF11 0xCB
#define M_SOF13 0xCD
#define M_SOF14 0xCE
#define M_SOF15 0xCF
#define M_SOI   0xD8            /* Start Of Image (beginning of datastream) */
#define M_EOI   0xD9            /* End Of Image (end of datastream) */
#define M_SOS   0xDA            /* Start Of Scan (begins compressed data) */
#define M_APP0  0xE0            /* Application-specific marker, type N */
#define M_APP12 0xEC            /* (we don't bother to list all 16 APPn's) */
#define M_COM   0xFE            /* COMment */


/*
 * Find the next JPEG marker and return its marker code.  We expect at least
 * one FF byte, possibly more if the compressor used FFs to pad the file.
 *
 * There could also be non-FF garbage between markers.  The treatment of such
 * garbage is unspecified; we choose to skip over it but emit a warning msg.
 * NB: this routine must not be used after seeing SOS marker, since it will not
 * deal correctly with FF/00 sequences in the compressed image data...
 */

static int
next_marker (void)
{
  int c;
  int discarded_bytes = 0;

  /* Find 0xFF byte; count and skip any non-FFs. */
  c = read_1_byte ();
  if (c == EOF)
    return (EOF);
  while (c != 0xFF)
    {
      discarded_bytes++;
      c = read_1_byte ();
      if (c == EOF)
        return (EOF);
    }
  /* Get marker code byte, swallowing any duplicate FF bytes.  Extra FFs
   * are legal as pad bytes, so don't count them in discarded_bytes. */
  do
    {
      c = read_1_byte ();
      if (c == EOF)
        return (EOF);
    }
  while (c == 0xFF);

  if (discarded_bytes != 0)
    fprintf (stderr, "jpgcom warning: garbage data found in JPEG file\n");
  return (c);
}


/*
 * Read the initial marker, which should be SOI.  For a JFIF file, the first
 * two bytes of the file should be literally 0xFF M_SOI.  To be more general,
 * we could use next_marker, but if the input file weren't actually JPEG at
 * all, next_marker might read the whole file and then return a misleading
 * error message...
 */

static int
first_marker (void)
{
  int c1, c2;

  c1 = NEXTBYTE ();
  if (c1 == EOF)
    ERR_RET ("non-JPEG");
  c2 = NEXTBYTE ();
  if ((c1 == EOF) || (c1 != 0xFF) || (c2 != M_SOI))
    ERR_RET ("non-JPEG");
  return (c2);
}


/*
 * Most types of marker are followed by a variable-length parameter segment.
 * This routine skips over the parameters for any marker we don't otherwise
 * want to process.  Note that we MUST skip the parameter segment explicitly in
 * order not to be fooled by 0xFF bytes that might appear within the parameter
 * segment; such bytes do NOT introduce new markers.
 */

static int
skip_variable (void)
/* Skip over an unknown or uninteresting variable-length marker */
{
  int rc;
  int length;

  /* Get the marker parameter length count */
  /* Length includes itself, so must be at least 2 */
  length = read_2_bytes ();
  if (length < 2)
    ERR_RET ("Bad-marker-length");
  length -= 2;
  /* Skip over the remaining bytes */
  rc = fseek (jpgfile, (long) length, 1);
  if (rc < 0)
    ERR_RET ("Bad-marker-data");
  return (0);
}


/*
 * Process a COM marker.  We want to print out the marker contents as legible
 * text; we must guard against non-text junk.
 */

static int
process_COM (void)
{
  int length;
  int ch;

  /* Get the marker parameter length count */
  /* Length includes itself, so must be at least 2 */
  length = read_2_bytes ();
  if (length < 2)
    ERR_RET ("Bad-COM-length");
  length -= 2;

  fprintf (stdout, "\t");
  while (length > 0)
    {
      ch = read_1_byte ();
      if (ch == EOF)
        ERR_RET ("Bad-COM-data");
      if (isprint (ch))
        putc (ch, stdout);
      else if (ch != 0)
        fprintf (stdout, "[%02x]", (unsigned int) ch);
      length--;
    }
  return (0);
}


/*
 * Parse the marker stream until SOS or EOI is seen; display any COM markers.
 * While the program wrjpgcom will always insert COM markers before SOFn, other
 * implementations might not, so we scan to SOS before stopping.
 */

static int
scan_JPEG_header (void)
{
  int marker, rc = 0;

  /* Expect SOI at start of file */
  if (first_marker () != M_SOI)
    return (EOF);

  /* Scan miscellaneous markers until we reach SOS. */
  for (;;)
    {
      marker = next_marker ();
      switch (marker)
        {
        case M_SOS:             /* stop before hitting compressed data */
          return (marker);

        case M_EOI:             /* in case it's a tables-only JPEG stream */
          return (marker);

        case M_COM:
          rc = process_COM ();
          if (rc == EOF)
            return (EOF);
          break;

        case M_APP12:
          fprintf (stdout, "\tAPP12");
          rc = process_COM ();
          if (rc == EOF)
            return (EOF);
          break;

        default:                   /* Anything else just gets skipped */
          rc = skip_variable ();   /* we assume it has a parameter count... */
          if (rc == EOF)
            return (EOF);
          break;
        }
    }                           /* end loop */
  return (0);
}


/*
 * A simple help function.
 */

static void
help (FILE *where)
{
  fprintf (where,
    "jpgcom " PACKAGE_VERSION "\n"
    "usage: jpgcom [-f file_list] [-h(elp)] [-p(ipe)] jpgfile...\n");
}

/*
 * Print the filename along with any comments.
 */

static void
print_com (char *filename)
{
  int rc;

  if ((jpgfile = fopen (filename, "rb")) == NULL)
    fprintf (stderr, "jpgcom error: can't open %s\n", filename);
  else
    {
      fprintf (stdout, "%s", filename);
      rc = scan_JPEG_header ();
      fprintf (stdout, "\n");
      (void) fclose (jpgfile);
    }
}


/*
 * The main program.
 */

int
main (int argc, char **argv)
{
  FILE *infile;
  char filename[PATH_MAX], *listname, *newline, *rc;
  int argn, opt, use_file, use_pipe;

  /* parse options */

  listname = "";
  use_file = 0;
  use_pipe = 0;
  while ((opt = getopt (argc, argv, "f:hp")) != OPTEND)
    switch (opt)
      {
      case 'f':
        use_file = 1;
        listname = optarg;
        break;
      case 'h':
        help (stdout);
        return (0);
      case 'p':
        use_pipe = 1;
        break;
      case '?':
        help (stderr);
        return (1);
      }
  if ((optind == argc) && (use_file == 0) && (use_pipe == 0))
    {
      help (stdout);
      return (0);
    }

  /* process files in listed in file specified by -f option */

  if (use_file)
    {
      infile = fopen (listname, "r");
      if (infile == NULL)
        fprintf (stderr, "jpgcom error: can't open %s!\n", listname);
      else
        {
          while (!feof (infile))
            {
              rc = fgets (filename, PATH_MAX - 1, infile);
              if (rc != NULL)
                {
                  newline = strchr (filename, '\n');
                  if (newline != NULL)
                    *newline = '\0';
                  if (strlen (filename) != 0)
                    print_com (filename);
                }
            }
          fclose (infile);
        }
    }

  /* process files listed on stdin (i.e., the -p option) */

  if (use_pipe)
    while (!feof (stdin))
      {
        rc = fgets (filename, PATH_MAX - 1, stdin);
        if (rc != NULL)
          {
            newline = strchr (filename, '\n');
            if (newline != NULL)
              *newline = '\0';
            if (strlen (filename) != 0)
              print_com (filename);
          }
      }


  /* process files given in the argument list */

  for (argn = 1; argn < argc; argn++)
    print_com (argv[argn]);

  /* All done. */

  return (0);
}
